/*
 * commands.h
 *
 * Copyright (C) 2012 - 2015 James Booth <boothj5@gmail.com>
 *
 * This file is part of Profanity.
 *
 * Profanity is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Profanity is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Profanity.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give permission to
 * link the code of portions of this program with the OpenSSL library under
 * certain conditions as described in each individual source file, and
 * distribute linked combinations including the two.
 *
 * You must obey the GNU General Public License in all respects for all of the
 * code used other than OpenSSL. If you modify file(s) with this exception, you
 * may extend this exception to your version of the file(s), but you are not
 * obligated to do so. If you do not wish to do so, delete this exception
 * statement from your version. If you delete this exception statement from all
 * source files in the program, then also delete it here.
 *
 */

#ifndef COMMANDS_H
#define COMMANDS_H

#include "ui/win_types.h"

// Command help strings
typedef struct cmd_help_t {
    const gchar *tags[20];
    const gchar *synopsis[50];
    const gchar *desc;
    const gchar *args[50][2];
    const gchar *examples[10];
} CommandHelp;

/*
 * Command structure
 *
 * cmd - The command string including leading '/'
 * func - The function to execute for the command
 * parser - The function used to parse arguments
 * min_args - Minimum number of arguments
 * max_args - Maximum number of arguments
 * help - A help struct containing usage info etc
 */
typedef struct cmd_t {
    gchar *cmd;
    gboolean (*func)(ProfWin *window, const char * const command, gchar **args);
    gchar** (*parser)(const char * const inp, int min, int max, gboolean *result);
    int min_args;
    int max_args;
    void (*setting_func)(void);
    CommandHelp help;
} Command;

gboolean cmd_execute_alias(ProfWin *window, const char * const inp, gboolean *ran);
gboolean cmd_execute_default(ProfWin *window, const char * inp);

gboolean cmd_about(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_account(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_autoaway(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_autoconnect(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_autoping(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_away(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_beep(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_caps(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_chat(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_chlog(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_clear(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_close(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_connect(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_decline(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_disco(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_disconnect(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_dnd(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_flash(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_gone(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_grlog(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_group(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_help(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_history(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_carbons(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_receipts(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_info(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_intype(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_invite(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_invites(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_join(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_leave(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_log(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_msg(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_nick(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_notify(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_online(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_otr(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_pgp(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_outtype(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_prefs(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_priority(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_quit(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_reconnect(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_room(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_rooms(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_bookmark(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_roster(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_software(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_splash(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_states(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_status(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_statuses(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_sub(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_theme(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_tiny(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_titlebar(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_vercheck(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_who(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_win(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_wins(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_xa(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_alias(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_xmlconsole(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_ping(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_form(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_occupants(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_kick(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_ban(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_subject(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_affiliation(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_role(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_privileges(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_presence(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_wrap(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_time(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_resource(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_inpblock(ProfWin *window, const char * const command, gchar **args);
gboolean cmd_encwarn(ProfWin *window, const char * const command, gchar **args);

gboolean cmd_form_field(ProfWin *window, char *tag, gchar **args);

#endif
