/*
 * chat_session.h
 *
 * Copyright (C) 2012 - 2015 James Booth <boothj5@gmail.com>
 *
 * This file is part of Profanity.
 *
 * Profanity is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Profanity is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Profanity.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give permission to
 * link the code of portions of this program with the OpenSSL library under
 * certain conditions as described in each individual source file, and
 * distribute linked combinations including the two.
 *
 * You must obey the GNU General Public License in all respects for all of the
 * code used other than OpenSSL. If you modify file(s) with this exception, you
 * may extend this exception to your version of the file(s), but you are not
 * obligated to do so. If you do not wish to do so, delete this exception
 * statement from your version. If you delete this exception statement from all
 * source files in the program, then also delete it here.
 *
 */

#ifndef CHAT_SESSION_H
#define CHAT_SESSION_H

#include <glib.h>

typedef struct chat_session_t {
    char *barejid;
    char *resource;
    gboolean resource_override;
    gboolean send_states;

} ChatSession;

void chat_sessions_init(void);
void chat_sessions_clear(void);

void chat_session_resource_override(const char * const barejid, const char * const resource);
ChatSession* chat_session_get(const char * const barejid);

void chat_session_recipient_active(const char * const barejid, const char * const resource,
    gboolean send_states);
void chat_session_recipient_typing(const char * const barejid, const char * const resource);
void chat_session_recipient_paused(const char * const barejid, const char * const resource);
void chat_session_recipient_gone(const char * const barejid, const char * const resource);
void chat_session_recipient_inactive(const char * const barejid, const char * const resource);

void chat_session_remove(const char * const barejid);

#endif
